#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright(C) 2006 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
conntracking.py is a part of pyctd which manages data extraction from
Netfilter Conntrack in a thread.
"""
import thread, time, logging
from xmlrpclib import DateTime
from socket import ntohs
from pynetfilter_conntrack import NetfilterConntrack, NFCT_DIR_MAX, IPS_NAMES, IPPROTO_NAMES, IPPROTO_ICMP
from auto_daemon import AutoDaemon

class DataCollector(object):
    #TODO: manage mutex on __current more deeply
    SLEEP_DURATION = 1
    
    def __init__(self):
        self.log = logging.getLogger(AutoDaemon.APP_NAME)
        self.__running = True
        self.__running_lock = thread.allocate_lock()
        self.__netfilter = NetfilterConntrack()
        self.__current = None
        self.__current_lock = thread.allocate_lock()
        self.__server_lock = thread.allocate_lock()
        self.__stopping_lock = thread.allocate_lock()
        thread.start_new_thread(self.run, ())
        #self.run() # debugging purpose only
        
    def __del__(self):
        self.stop()
        
    def stop(self):
        self.__setRunning(False)
        # Waiting for the other thread
        self.__stopping_lock.acquire()
        self.__stopping_lock.release()
    
    def run(self):
        self.__stopping_lock.acquire()
        while (self.__getRunning()):
            self.__server_lock.acquire()
            try:
                self.__refresh()
            finally:
                # Let the server's thread killing connections
                self.__server_lock.release()
            time.sleep(self.SLEEP_DURATION)
        self.__stopping_lock.release()
    
    def kill(self, conn_id):
        self.__server_lock.acquire() # Synchronize with data's thread
        try:
            connset = self.__getCurrent()
            for id in conn_id:
                conn = connset.dict.get(str(id))
                if conn is None:
                    continue
                self.log.info("Deleting conntrack: %s" %(conn))
                self.__netfilter.delete_conntrack(conn.conntrack)
            self.__refresh()
        finally:
            self.__server_lock.release()
    
    def change(self, field, conn_id, value):
        self.__server_lock.acquire()
        try:
            self.__refresh()
            connset = self.__getCurrent()
            conn = connset.dict.get(str(conn_id))
            if conn is not None:
                old_val = None
                try:
                    if field == "mark":
                        old_val = conn.conntrack.mark
                        conn.conntrack.mark = long(value)
                    elif field == "timeout":
                        old_val = conn.conntrack.timeout
                        conn.conntrack.timeout = long(value)
                    else:
                        raise Exception("%s fields cannot be changed.")
                except TypeError:
                    raise Exception("Type of the new value unmatched (%s is a %s ; %s is a %s)" %(field, type(old_val), value, type(value)))
                self.__netfilter.update_conntrack(conn.conntrack)
                self.log.info("Replacing %s from '%s' to '%s' in %s" %(field, old_val, value, conn))
                self.__refresh()
        finally:
            self.__server_lock.release()
    
    def __refresh(self):
        conn = ConnectionSet(self.__netfilter.create_table(), time.time())
        conn.calculate(self.__getCurrent())
        self.__setCurrent(conn)
    
    def __getRunning(self):
        self.__running_lock.acquire()
        running = self.__running
        self.__running_lock.release()
        return running
    def __setRunning(self, flag):
        self.__running_lock.acquire()
        self.__running = flag
        self.__running_lock.release()
    
    def __getCurrent(self):
        self.__current_lock.acquire()
        current = self.__current
        self.__current_lock.release()
        return current
    def __setCurrent(self, new):
        self.__current_lock.acquire()
        self.__current = new
        self.__current_lock.release()
    current = property(__getCurrent)

    def __str__(self):
        return '<DataCollector %s>' %(self.__getCurrent())

class ConnectionSet:
    #TODO export into XML
    def __init__(self, table, hit):
        self.dict = {}
        self.hit = hit
        for conn in table:
            self.dict[str(conn.id)] = Connection(conn, hit)
    
    def calculate(self, ref):
        if ref is not None:
            for (id,conn) in self.dict.items():
                ref_conn = ref.dict.get(id)
                if ref_conn is not None:
                    conn.setByteRate(ref_conn)
    
    def getStruct(self):
        root = {}
        root["timestamp"] = DateTime(self.hit)
        conntab = []
        keys = self.dict.keys()
        keys = map(int, keys)
        keys.sort()
        keys = map(str, keys)
        for key in keys:
            conn = self.dict[key]
            conntrack = {}
            conntrack["id"] = int(conn.conntrack.id)
            conntrack["mark"] = str(conn.conntrack.mark) #TODO fixme: use int() cast
            conntrack["use"] = bool(conn.conntrack.use)
            conntrack["timeout"] = int(conn.conntrack.timeout)
            stattab = []
            flags = [ name for mask, name in IPS_NAMES if conn.conntrack.status & mask ]
            for flag in flags:
                stattab.append(str(flag))
            conntrack["status"] = stattab
            ttab = []
            for i in range(NFCT_DIR_MAX):
                t = conn.conntrack.tuple[i]
                tuple = {}
                tuple["l3protonum"] = int(t.l3protonum)
                tuple["protonum"] = str(IPPROTO_NAMES[t.protonum])
                tuple["src"] = str(t.src.getIP(t.l3protonum))
                tuple["dst"] = str(t.dst.getIP(t.l3protonum))
                if t.protonum == IPPROTO_ICMP:
                    tuple["l4src"] = int(t.l4src.icmp.id)
                    tuple["l4dst"] = int(t.l4dst.icmp.type)
                else:
                    tuple["l4src"] = int(ntohs(t.l4src.getValue(t.protonum)))
                    tuple["l4dst"] = int(ntohs(t.l4dst.getValue(t.protonum)))
                tuple["packets"] = int(conn.conntrack.counters[i].packets)
                tuple["bytes"] = int(conn.conntrack.counters[i].bytes)
                tuple["byterate"] = float(conn.byterate[i])
                ttab.append(tuple)
            conntrack["tuples"] = ttab
            conntab.append(conntrack)
        root["table"] = conntab
        return root
    
    def __str__(self):
        s = "Connection Set at %s:\n" %(self.hit)
        for conn in self.dict.values():
            s = "%s\t%s\n" %(s, conn)
        return s


class Connection:
    def __init__(self, conntrack, hit):
        self.conntrack = conntrack
        self.byterate = [0.0, 0.0]
        self.hit = hit
    
    def setByteRate(self, ref_conn):
        #TODO: calculate an average or immediate byterate ?
        ref_hit = ref_conn.hit
        for i in range(NFCT_DIR_MAX):
            current_bytes = self.conntrack.counters[i].bytes
            ref_bytes = ref_conn.conntrack.counters[i].bytes
            self.byterate[i] = float(current_bytes - ref_bytes) / float(self.hit - ref_hit)
    
    def __str__(self):
        return "<Connection %(conntrack)s %(byterate)s>" %{
                                                'conntrack': self.conntrack,
                                                'byterate': self.byterate,
                                                'hit': self.hit }
    