#!/usr/bin/env python
# -*- coding: utf-8 -*-

from SimpleXMLRPCServer import SimpleXMLRPCServer
import logging, thread, time, random, sha
from auto_daemon import AutoDaemon
"""
Copyright(C) 2006 INL
Written by Damien Boucard <damien.boucard AT inl.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
---
server.py is a part of pyctd which manages XML-RPC service.
"""
class PyctServer(SimpleXMLRPCServer):
    allow_reuse_address = True
    SESSION_DURATION = 10 # seconds
    
    def __init__(self, data, host, port, mod_auth):
        self.log = logging.getLogger(AutoDaemon.APP_NAME)
        self.data = data
        if mod_auth is not None:
            self.authenticate = mod_auth.authenticate
        else:
            self.authenticate = self.__authenticate
        self.__sessions = {}
        self.__session_lock = thread.allocate_lock()
        SimpleXMLRPCServer.__init__(self, (host, port), logRequests = False)
    
    def _dispatch(self, method, params):
        self.update_sessions()
        if method == "login":
            if self.authenticate(*params):
                result = self.create_session(*params)
                self.log.info("'%s' is successfully authenticated" %(params[0]))
            else:
                self.log.error("Authentication of %s failed" %(params[0]))
        else:
            params = list(params)
            id = params.pop(0)
            if id not in self.__sessions:
                raise Exception('Wrong id or session has timed out.')
            try:
                # We are forcing the 'rpc_' prefix on methods that are
                # callable through XML-RPC to prevent potential security
                # problems
                func = getattr(self, 'rpc_' + method)
            except AttributeError:
                self.log.error('Method "%s" is not supported' % method)
                raise Exception('Method "%s" is not supported' % method)
            try:
                result = func(*params)
            except:
                self.log.exception("Unhandled exception while serving %s%s" %(func.__name__, tuple(params)))
                raise Exception('Internal server error')
            self.log.info('Method "%s%s" is successfully called' %(method, tuple(params)))
        return result
    
    def __authenticate(self, *params):
        return True
    
    def create_session(self, username, password):
        self.__session_lock.acquire()
        try:
            id = sha.new(str(random.random())).hexdigest()
            while id in self.__sessions:
                id = sha.new(str(random.random())).hexdigest()
            self.__sessions[id] = time.time() + self.SESSION_DURATION
        finally:
            self.__session_lock.release()
        return id
    
    def update_sessions(self):
        t = time.time()
        self.__session_lock.acquire()
        try:
            for (id, timeout) in self.__sessions.copy().items():
                if timeout < t:
                    del self.__sessions[id]
        finally:
            self.__session_lock.release()
    
    def rpc_view(self):
        return self.data.current.getStruct()
    
    def rpc_kill(self, conn_id):
        self.data.kill(conn_id)
        return 0
    
    def rpc_change(self, field, conn_id, value):
        self.data.change(field, conn_id, value)
        return 0
